package eGuideDog::Browser::HTML;

use strict;
use warnings;
use vars qw(@ISA);
require eGuideDog::Browser::Widget;
@ISA = qw(eGuideDog::Browser::Widget);
use threads;
use threads::shared; # the position of this definition is very important!!
use Encode qw/encode decode/;
use HTML::Element;
use eGuideDog::Browser::Const;
use eGuideDog::Browser::HTML::Navigate;
use eGuideDog::Browser::InputText;
use eGuideDog::Browser::InputPassword;
use eGuideDog::Browser::HTML::Format;
use LWP::UserAgent;
use eGuideDog::Basic;
#use Curses; # This is for key mapping. It's not good to port it to Windows.

sub new;
sub navigate;
sub progress;
sub percent;
sub load; # (url_or_tree)
sub reload; # (encoding)
sub get_html; # get raw html (url, postdata, cookie_file)
sub get_tree; # (content)
sub get_page; # (url) => DOM tree
sub get_abs_url;
sub gen_buffer;
sub gen_submit_content;
sub press;
sub browser;

sub new {
  my ($class, $url) = @_;
  my $self = {index => 0,
	      url => undef,
	      url_base => undef,
	      encoding => 'utf8',
	      valid_nodes => 0,
#	      agent => eGuideDog::Browser::UserAgent->new(),
	      screen => undef, # current screen buffer
#	      scr_bufs => [],
	      browser => undef,
	      root => undef,
	      navigate => undef,
	      content => undef, # html raw content buffer
	     };
  my $percent : shared = 0;
  my $progress : shared = 0;
  my $speed :shared = 0;
  my $eta : shared = 0;
  my $finished : shared = 0;
  my $new_url : shared = '';
  $self->{percent} = \$percent;
  $self->{progress} = \$progress;
  $self->{speed} = \$speed;
  $self->{eta} = \$eta;
  $self->{finished} = \$finished;
  $self->{new_url} = \$new_url; # For record redirected URL
  bless $self, __PACKAGE__;
  if ($url) {
    $self->load($url);
  }
  return $self;
}

sub navigate {
  my $self = shift;
  return $self->{navigate};
}

sub url {
  my ($self, $url) = @_;

  if ($url) {
    $self->{url} = $url;
  } else {
    if (${$self->{new_url}}) {
      return ${$self->{new_url}};
    } else {
      return $self->{url};
    }
  }
}

sub browser {
  my ($self, $browser) = @_;
  $self->{browser} = $browser;
  return $self->{browser};
}

sub progress {
  my $self = shift;
  return ${$self->{progress}};
#  return ${$self->{agent}->{progress}};
}

sub percent {
  my $self = shift;
  if (${$self->{percent}}) {
    return ${$self->{percent}};
  } else {
    return undef;
  }
#  if (${$self->{agent}->{content_len}}) {
#    return sprintf("%2.1f", ${$self->{agent}->{progress}} * 100 / ${$self->{agent}->{content_len}});
#  } else {
#    return undef;
#  }
}

sub speed {
  my $self = shift;
  if (${$self->{speed}}) {
    return ${$self->{speed}};
  } else {
    return undef;
  }
}

sub eta {
  my $self = shift;
  if (${$self->{eta}}) {
    return ${$self->{eta}};
  } else {
    return undef;
  }
}

sub finished {
  my $self = shift;
  return ${$self->{finished}};
#  return ${$self->{agent}->{finished}};
}

sub get_abs_url {
  my ($self, $url) = @_;

  if ($url !~ /^http:\/\// && $url !~ /^https:\/\//) { # relative url
    my $base = $self->{url_base};
    my $protocol;
    my $site_base;
    if ($self->{url_base} =~ /^http:\/\//) {
      $base =~ s/^http:\/\/([^\/]*)//;
      $site_base = 'http://' . $1;
      $protocol = 'http://';
    } else {
      $base =~ s/^https:\/\/([^\/]*)//;
      $site_base = 'https://' . $1;
      $protocol = 'https://';
    }
    $base =~ s/[^\/]*$//;

    if ($base eq '') {
      $base = $site_base . '/';
    } else {
      $base = $site_base . $base;
    }

    if ($url =~ /^\//) { # website relative
      $url = $site_base . $url;
    } else { # directory relative
      $url = $base . $url;
    }
  }
  return $url;
}

sub reload {
  my ($self, $encoding) = @_;

  $self->{encoding} = $encoding;
  my $format = eGuideDog::Browser::HTML::Format->new();
  my $tree = $format->filter_page($self->get_tree(\$self->{content}));
  my ($screen_buffer, $speech_buffer) = $self->gen_buffer();
  my $scr_buf = eGuideDog::Terminal::Buffer->new($screen_buffer);
  my $spe_buf = eGuideDog::Terminal::Buffer->new($speech_buffer);
  $self->screen($scr_buf);
  $self->speechs($spe_buf);
  $self->{navigate}->restart();
}

sub load {
  my ($self, $url_or_tree) = @_;

  my ($url, $tree);
  if (ref($url_or_tree) eq 'HTML::Element') {
    $tree = $url_or_tree;
  } else {
    $url = $url_or_tree;
    $tree = $self->get_page(get_full_path($url_or_tree));
  }
  my $format = eGuideDog::Browser::HTML::Format->new();
  if ($tree) {
    $tree = $format->filter_page($tree);
  } else {
    return undef;
  }

  if (defined $url) {
    $self->{url} = $url;
  }
  $self->{root} = $tree;
  $self->{navigate} = eGuideDog::Browser::HTML::Navigate->new($tree);

  # url base
  my @bases = $self->{navigate}->{root}->look_down('_tag', 'base');
  if ($bases[0] && $bases[0]->attr('href')) {
    if ($bases[0]->attr('href') =~ /^http/) {
      $self->{url_base} = $bases[0]->attr('href');
    } else {
      $self->{url_base} = 'http://' . $bases[0]->attr('href');
    }
  } else {
    my $url;
    if (${$self->{new_url}}) {
      $url = ${$self->{new_url}};
    } else {
      $url = $self->{url};
    }
    if ($url =~ /^http/) {
      $self->{url_base} = $url;
    } else {
      $self->{url_base} = 'http://' . $url;
    }
  }

  # get encoding
  my @metas = $self->{root}->look_down('_tag', 'meta'); # search only <head> will improve speed.
  foreach my $meta (@metas) {
    my $http_equiv = lc($meta->attr('http-equiv'));
    if ($http_equiv eq 'content-type') {
      my $content = lc($meta->attr('content'));
      $content =~ /charset=(.*)/;
      my $encoding = Encode::resolve_alias($1);
      if ($encoding) {
	$self->{encoding} = $encoding;
      }
    }
  }

  my ($screen_buffer, $speech_buffer) = $self->gen_buffer();
  my $scr_buf = eGuideDog::Terminal::Buffer->new($screen_buffer);
  my $spe_buf = eGuideDog::Terminal::Buffer->new($speech_buffer);
#  push(@{$self->{scr_bufs}}, eGuideDog::Terminal::Buffer->new($screen_buffer));
  $self->screen($scr_buf);
  $self->speechs($spe_buf);
  $self->{navigate}->restart();

  return $self;
}

# input: URL
# return: HTML raw content
sub get_html {
  my ($self, $url, $postdata, $cookie_file) = @_;

  my $post_opt = '';
  $post_opt = "--post-data='$postdata'" if ($postdata);

  if (not $cookie_file) {
    $cookie_file = $self->{browser}->cookie_file;
  }
  my $cookie_opt = '';
  $cookie_opt = "--keep-session-cookies --save-cookies $cookie_file --load-cookies $cookie_file" if ($cookie_file);

  my $log = gen_tmpfile('wget');
  open(WGET_LOG, '>', $log) || die "Fail to open $log";
  close(WGET_LOG) || warn "Fail to close $log";
  my $file = gen_tmpfile('file');

  my $result = `export LANG=en_US.UTF-8; export LANGUAGE=en_US.UTF-8; wget -c -o $log -b --progress=bar:force -O $file $post_opt $cookie_opt $url`;
  $result =~ /.*\s(\d*)[.]/;
  my $pid = $1;
  my $finished = 0;
  my $error = '';
  if ($pid) {
    open(WGET_LOG, '<', $log) || die "Fail to open $log";
    while ((not $finished) && ($error eq '')) {
      my $curpos = tell(WGET_LOG); # tell will fail sometimes, does take it as error. # warn "Fail to tell $log";
      while(<WGET_LOG>) {
	if (/^\r(...) \[.*\] ([0-9,]*)\s*([0-9.,-MKB]*\/s)\s*(.*)/) {
	  # update progress
	  my $percent = $1;
	  my $progress = $2;
	  my $speed = $3;
	  my $eta = $4;
	  ${$self->{percent}} = $percent if ($percent ne '   ');
	  ${$self->{progress}} = $progress if ($progress);
	  ${$self->{speed}} = $speed if ($speed eq '--.--K/s');
	  ${$self->{eta}} = $eta if ($eta);
	} elsif (/$file' saved/) {
	  $finished = 1;
	} elsif (/ERROR 404: Not Found./) {
	  $error = gettext('ERROR 404: Not Found.');
	} elsif (/failed: Temporary failure in name resolution./) {
	  $error = gettext('failed: Temporary failure in name resolution.');
	} elsif (/Invalid host name./) {
	  $error = gettext('Invalid host name.');
	} elsif (/Location: ([^ ]*) \[following\]/) {
	  # redirect, the previous line should be:
	  # HTTP request sent, awaiting response... 302 Found
	  ${$self->{new_url}} = $1;
#	  warn "redirect: $1";
	}

	$curpos = tell(WGET_LOG);
      }
      sleep(1);
      seek(WGET_LOG, $curpos, 0) || die "Fail to seed $log";
    }
    close(WGET_LOG) || warn "Fail to close $log";
  } else {
    warn "Fail to get $url: $result";
    return "Fail to get $url";
  }

  if ($finished) {
    ${$self->{finished}} = 1;
    open(WGET_LOG, '<', $file) || die "Fail to open $file";
    read(WGET_LOG, $self->{content}, -s $file) || die "Fail to read $file";
    close(WGET_LOG) || warn "Fail to close $file";
    return $self->{content};
  } else { # error
    ${$self->{finished}} = 1;
    return $error;
  }
}

# load from html content and generate a tree
# content is a reference of html text
#
##### This method should be deprecated! #####
# There will be problem with URL redirection, new url not known!
#
sub get_tree {
  my ($self, $content) = @_;
  my $tree;
#  if ($$content =~ /<html>.*<\/html>/) { # maybe low speed
    $tree = HTML::TreeBuilder->new_from_content($$content);
#  } else {
#    $tree = HTML::TreeBuilder->new_from_content('<html><body>' . $$content . '</body></html>');
#  }
  die if (not defined $tree);
  bless $tree, 'HTML::Element';
  return $tree;
}

# get html page
# input: $url
# output: an html DOM tree
sub get_page {
  my ($self, $url) = @_;

  my $html;
  my $request = HTTP::Request->new(GET => $url);
  my $agent = LWP::UserAgent->new();
  my $response = $agent->request($request);
  if ($response->is_success) {
    $html = $response->content;
  } else {
    return undef;
  }
  my $tree = HTML::TreeBuilder->new_from_content($html);

  bless $tree, 'HTML::Element';
  return $tree;
}

sub gen_buffer {
  my ($self) = @_;

  my @node_buffer;
  my @screen_buffer;
  my @speech_buffer;
  my $screen;
  my $speech;

  my $navigate = $self->{navigate};
  $self->{valid_nodes} = 0;

  while ($navigate->moveto_next_valid_node()) {
    if (ref $navigate->{current_node}) {
      # title
      if ($navigate->{current_node}->tag eq 'title') {
	$screen = '<title>' . $navigate->{current_node}->as_trimmed_text;
	$speech = 'Title: ' . $navigate->{current_node}->as_trimmed_text;
      }

      # frame, iframe
      elsif ($navigate->{current_node}->tag eq 'iframe'
	     || $navigate->{current_node}->tag eq 'frame') {
	my $desc = $navigate->{current_node}->attr('longdesc');
	$desc = $navigate->{current_node}->attr('title') if (!$desc);
	if ($desc) {
	  $screen = '<' . $navigate->{current_node}->tag . ' src="' . $navigate->{current_node}->attr('src') . '">' . $desc;
	  $speech = 'a frame about ' . $desc . ', which links to ' . $navigate->{current_node}->attr('src');
	} else {
	  $screen = '<' . $navigate->{current_node}->tag . '>' . $navigate->{current_node}->attr('src');
#	  my $url = $navigate->{current_node}->attr('src');
#	  $url =~ /^(http:\/\/)?([^\/]*)/;
	  $speech = 'a frame links to ' . $navigate->{current_node}->attr('src');
	}
      }

      # link
      elsif ($navigate->{current_node}->tag eq 'a') {
	my $href = $navigate->current_node->attr('href');
	$href = '' if (not defined $href);
	$screen = '<a href=' . $href . '>' . $navigate->{current_node}->as_trimmed_text;
	$speech = 'Link: ' . $navigate->{current_node}->as_trimmed_text;
      }

      # sitemap
      elsif ($navigate->{current_node}->tag eq 'img'
	     && $navigate->{current_node}->attr('usemap')) {
	my $mapname = $navigate->{current_node}->attr('usemap');
	$mapname =~ s/^#//;
	my @maps = $navigate->{current_node}->root->look_down('_tag', 'map',
							 'name', $mapname);
	my $num = 0;
	$num = @{$maps[0]->content_array_ref} if ($maps[0]);
	$screen = '<img usemap=' . $navigate->{current_node}->attr('usemap') . '> ' . $num . ' links';

	$speech = 'a site map with ' . $num . ' links.';
      }

      # select
      elsif ($navigate->{current_node}->tag eq 'select') {
	my $len = 0;
	my $multiple = $navigate->{current_node}->attr('multiple');
	my $size = $navigate->{current_node}->attr('size');
	my $selections = 0;
	my $select_id = 0;

	# load options
	foreach my $option_node ($navigate->{current_node}->content_refs_list) {
	  $option_node = $$option_node;
	  next if ($option_node->tag ne 'option');
	  if ($option_node->attr('value')) {
	    $len++;
	    if ($option_node->attr('selected')) {
	      $selections++;
	      $select_id = $option_node->idf();
	    }
	  }
	}
	if ($multiple) {
	  if ($size) {
	    $screen = '<select size="' . $size . '">' . $len . ' optinos, ' . $selections . ' selected.';
	    $speech = 'A selection with ' . $len . ' options which ' . $selections . ' have been selected while at most ' . $size . ' selections can be selected.';
	  } else {
	    $screen = '<select>' . $len . ' optinos, ' . $selections . ' selected.';
	    $speech = 'A selection with ' . $len . ' options which ' . $selections . ' have been selected.';
	  }
	} else {
	  if ($select_id) {
	    $screen = '<select>' . $len . ' options, ' . $navigate->{current_node}->look_down('id', $select_id)->as_trimmed_text . ' selected.';
	    $speech = 'a selection with ' . $len . ' options. ' . $navigate->{current_node}->look_down('id', $select_id)->as_trimmed_text . ' has been selected.';
	  } else {
	    $screen = '<select>' . $len . ' options, no option selected.';
	    $speech = 'a selection with ' . $len . ' options. No option has selected yet.';
	  }
	}
      }

      # ol (ordered lists)
      elsif ($navigate->{current_node}->tag eq 'ol') {
#	my $len = 0;
#	my $ol_index = $self->{valid_nodes};
#	push(@node_buffer, $navigate->{current_node});
#	push(@screen_buffer, '');
#	push(@sound_buffer, ''); # reserve for 'ol'
#	$self->{valid_nodes}++;
#	foreach my $li_node ($navigate->{current_node}->content_refs_list) {
#	  $li_node = $$li_node;
#	  next if ($li_node->tag ne 'li');
#	  $len++;
#	  my $text = $li_node->as_trimmed_text;
#	  if (!Encode::is_utf8($text)) {
#	    $text = decode($navigate->{encoding}, $text);
#	  }
#	  $screen = '<li>' . $text;
#	  $sound = 'Item ' . $len . ': ' . $text;
#	  push(@node_buffer, $li_node);
#	  push(@screen_buffer, $screen);
#	  push(@sound_buffer, $sound);
#	  $self->{valid_nodes}++;
#	}

	# refilled <ol>
	my $len = @{$navigate->{current_node}->content_array_ref};
	$screen = '<ol>' . $len . ' items';
	$speech = 'an ordered list with ' . $len . ' items.';
#	$screen_buffer[$ol_index] = '<ol>' . $len . ' items';
#	$sound_buffer[$ol_index] = 'an ordered list with ' . $len . ' items.';
#	next;
      }

      # ul (unordered lists)
      elsif ($navigate->{current_node}->tag eq 'ul') {
#	my $len = 0;
#	my $ol_index = $self->{valid_nodes};
#	push(@node_buffer, $navigate->{current_node});
#	push(@screen_buffer, '');
#	push(@sound_buffer, ''); # reserve for 'ul'
#	$self->{valid_nodes}++;
#	foreach my $li_node ($navigate->{current_node}->content_refs_list) {
#	  $li_node = $$li_node;
#	  next if ($li_node->tag ne 'li');
#	  $len++;
#	  my $text = $li_node->as_trimmed_text;
#	  if (!Encode::is_utf8($text)) {
#	    $text = decode($navigate->{encoding}, $text);
#	  }
#	  $screen = '<li>' . $text;
#	  $sound = 'Item: ' . $text;
#	  push(@node_buffer, $li_node);
#	  push(@screen_buffer, $screen);
#	  push(@sound_buffer, $sound);
#	  $self->{valid_nodes}++;
#	}

	# refilled <ul>
	my $len = @{$navigate->{current_node}->content_array_ref};
	$screen = '<ul>' . $len . ' items';
	$speech = 'List with ' . $len . ' items.';
#	$screen_buffer[$ol_index] = '<ul>' . $len . ' items';
#	$speech_buffer[$ol_index] = 'an unordered list with ' . $len . ' items.';
#	next;
      }

      # li(list item)
      elsif ($navigate->{current_node}->tag eq 'li') {
	if ($navigate->{current_node}->attr('type') eq 'e-guidedog_paragraph') {
	  my $text = $navigate->{current_node}->as_trimmed_text;
	  if (!Encode::is_utf8($text)) {
	    $text = decode($self->{encoding}, $text);
	  }
	  if ($navigate->{current_node}->parent->tag eq 'ul') {
	    $screen = '<li>' . $text;
	    $speech = 'Item: ' . $navigate->{current_node}->as_trimmed_text;
	  } elsif ($navigate->{current_node}->parent->tag eq 'ol') {
	    $screen = '<li ' . ($navigate->{current_node}->pindex + 1) . '>' . $text;
#	    $speech = 'Item ' . $navigate->{current_node}->pindex . ': ' . $text;
# wide char problem
	    $speech = 'Item ' . ($navigate->{current_node}->pindex + 1) . ': ' . $navigate->{current_node}->as_trimmed_text;
	  } else {
	    $screen = 'a bug: li without ol or ul';
	    $speech = $screen;
	  }
	} else {
	  next;
	}
      }

      # text input
      elsif ($navigate->{current_node}->tag eq 'input'
	     && (lc($navigate->{current_node}->attr('type')) eq 'text'
		 || !$navigate->{current_node}->attr('type'))) {
	my $value = $navigate->{current_node}->attr('value');
	$value = '' if (not defined $value);
	$screen = "<input type=text> $value";
	$speech = "Text input: $value";
      }

      # password input
      elsif ($navigate->{current_node}->tag eq 'input'
	     && (lc($navigate->{current_node}->attr('type')) eq 'password')) {
	my $value = $navigate->current_node->attr('value');
	$value = '' if (not defined $value);
	$value = '*' x length($value);
	$screen = '<input type=password>' . $value;
	$speech = "Password: $value";
      }

      # image input
      elsif ($navigate->{current_node}->tag eq 'input'
	     && lc($navigate->{current_node}->attr('type')) eq 'image') {
	$screen = '<input type=image>';
	$speech = 'a submit image';
      }

      # submit input
      elsif ($navigate->{current_node}->tag eq 'input'
	     && lc($navigate->{current_node}->attr('type')) eq 'submit') {
	$screen = '<input type=submit>';
	$speech = 'a submit button';
      }

      # div class='e-guidedog_paragraph'
      elsif ($navigate->{current_node}->tag eq 'div'
	     && $navigate->{current_node}->attr('class') eq 'e-guidedog_paragraph') {
	$screen = $navigate->{current_node}->as_trimmed_text;
	$speech = $screen;
      }

      # legend(caption of fieldset)
      elsif ($navigate->{current_node}->tag eq 'legend') {
	$screen = '<legend>' . $navigate->{current_node}->as_trimmed_text;
	$speech = 'A field of ' . $navigate->{current_node}->as_trimmed_text;
      }

      # object
      elsif ($navigate->{current_node}->tag eq 'object') {
	my $data = $navigate->{current_node}->attr('data');
	if ($data =~ /[.]html$/) {
	  $screen = '<object data="' . $data . '">';
	  $speech = 'An object source from ' . $data;
	}
	$screen = '<object>' . $navigate->{current_node}->as_trimmed_text;
	$speech = 'An object about ' . $navigate->{current_node}->as_trimmed_text;
      }

      # img with alt or longdesc
      elsif ($navigate->{current_node}->tag eq 'img'
	     && $navigate->{current_node}->attr('alt')) {
	if ($navigate->{current_node}->attr('longdesc')) {
	  $screen = '<img longdesc="' . $navigate->{current_node}->attr('longdesc') . '">' . $navigate->{current_node}->attr('alt');
	  $speech = 'An image about ' . $navigate->{current_node}->attr('alt') . '. There is a long description for this image.';
	} else {
	  $screen = '<img>' . $navigate->{current_node}->attr('alt');
	  $speech = 'An image about ' . $navigate->{current_node}->attr('alt');
	}
      }

      # data table
      elsif ($navigate->{current_node}->tag eq 'div'
	     && $navigate->{current_node}->attr('class') eq 'e-guidedog_table') {
	my $caption;
	my $td_num = 0;
	foreach my $table_node ($navigate->{current_node}->content_refs_list) {
	  if (ref $$table_node) {
	    if ($$table_node->tag eq 'table') {
	      foreach my $child_node ($$table_node->content_refs_list) {
		if (ref $$child_node) {
		  if ($$child_node->tag eq 'caption') {
		    $caption = $$child_node->as_trimmed_text;
		  } elsif ($$child_node->tag eq 'tr') {
		    foreach my $tr_child ($$child_node->content_refs_list) {
		      if (ref $$tr_child) {
			if ($$tr_child->tag eq 'td') {
			  $td_num++;
			}
		      }
		    }
		  }
		}
	      }
	      if (!$caption
		  && $$table_node->attr('summary')
		  && $$table_node->attr('summary') ne '') {
		$caption = $$table_node->attr('summary');
	      }
	    }
	  }
	}

	if ($caption) {
	  $screen = '<table>' . $td_num . ' cells about ' . $caption;
	  $speech = 'A data table with ' . $td_num . ' cells about' . $caption;
	} else {
	  $screen = '<table>' . $td_num . ' cells.';
	  $speech = 'A data table with ' . $td_num . ' cells.';
	}
      }

      # <div class='e-guidedog_part'>
      elsif ($navigate->{current_node}->tag eq 'div'
	     && $navigate->{current_node}->attr('class') eq 'e-guidedog_part') {
	$screen = $navigate->{current_node}->starttag;
	$speech = 'Part ' . $navigate->{current_node}->attr('part_no') . ' of ' . $navigate->{current_node}->attr('parts') . ', level ' . $navigate->{current_node}->attr('level') . ' with ' . $navigate->{current_node}->attr('quantity') . ' ' . $navigate->{current_node}->attr('tags') . ($navigate->{current_node}->attr('quantity') > 1 ? 's.' : '.');
	if (defined $navigate->{current_node}->attr('mainpart')) {
	  $speech .= ' This is the main part.';
	}
      }

      # unknown node
      else {
	$screen = $navigate->{current_node}->starttag;
	$speech = $screen;
      }
    } else {
      $screen = $navigate->{current_node};
      $screen =~ s/^\s+//;
      $screen =~ s/\s+$//; # trim
      $speech = $screen;
    }

    if (!Encode::is_utf8($screen)) {
      $screen = decode($self->{encoding}, $screen);
      $speech = decode($self->{encoding}, $speech);
    }
    $navigate->{current_node}->attr('e-guidedog_no', ++$self->{valid_nodes});
    push(@node_buffer, $navigate->{current_node});
    $screen = $self->{valid_nodes} . ': ' . $screen;
    push(@screen_buffer, $screen);
    push(@speech_buffer, $speech);
  }

  $navigate->{nodes} = \@node_buffer;
  $navigate->{valid_nodes} = $self->{valid_nodes};

  return (\@screen_buffer, \@speech_buffer);
}

sub gen_submit_content {
  my ($self, $form_node, $contents) = @_;

  $form_node->traverse (
    [ # Callbacks;
     # pre-order callback:
     sub {
       my $node = $_[0];
       return HTML::Element::OK if (!$node->attr('name'));

       if (!$node->attr('disabled')) {
	 # input text, hidden, submit, image
	 if ($node->tag eq 'input') {
	   my $type = lc($node->attr('type'));
	   $type = 'text' if (!$type);
	   if ($type eq 'text'
	       || ($type eq 'checkbox' and $node->attr('checked'))
	       || ($type eq 'radio' and $node->attr('checked'))
	       || $type eq 'password'
	       || $type eq 'image'
	       || $type eq 'submit'
	       || $type eq 'hidden') {
	     push(@{$contents}, $node->attr('name'), $node->attr('value'));
	   }
	 }

	 # select
	 elsif ($node->tag eq 'select') {
	   my @selected_nodes = $node->look_down('selected', defined);
	   foreach my $selected_node (@selected_nodes) {
	     push(@{$contents}, $node->attr('name'), $selected_node->attr('value'));
	   }
	 }
       }
       return HTML::Element::OK; # keep traversing
     },
     # post-order callback:
     undef
    ],
    1, # don't call the callbacks for text nodes
  );
#  warn @{$contents};
}

sub press {
  my ($self, $key) = @_;

  my $c = eGuideDog::Browser::Const::new();

  # down2
  if ($key eq $c->KEYS_DOWN) {
    #    my $node = $self->{navigate}->moveto_next_valid_node(); # use buffer to make faster
    if ($self->{navigate}->{node_index} < $#{$self->{navigate}->{nodes}}) {
      $self->{navigate}->{node_index}++;
      $self->screen->index($self->screen->index + 1);
      $self->speechs->index($self->speechs->index + 1);
#      $self->speech($self->screen->lines->[$self->screen->index]);
    } else {
      my $command = $self->{browser}->command;
      if ($command && $command == eGuideDog::Browser::Const::AUTO_SPEAK) {
	$self->{browser}->command('');
      }
      $self->speech(gettext('No more contents'));
    }
  }

  # up
  elsif ($key eq $c->KEYS_UP) {
    if ($self->{navigate}->{node_index} > 0) {
      $self->{navigate}->{node_index}--;
      $self->screen->index($self->screen->index - 1);
      $self->speechs->index($self->speechs->index - 1);
#      $self->speech($self->screen->lines->[$self->screen->index]);
    } else {
      $self->speech(gettext('No more contents'));
    }
  }

  # pageup
  elsif ($key eq $c->KEYS_PAGEUP) {
    if ($self->{navigate}->moveto_previous_part()) {
      $self->screen->index($self->{navigate}->{node_index});
      $self->speechs->index($self->{navigate}->{node_index});
#      $self->speech($self->screen->lines->[$self->screen->index]);
    } else {
      $self->speech(gettext('No more parts'));
    }
  }

  # pagedown
  elsif ($key eq $c->KEYS_PAGEDOWN) {
    if ($self->{navigate}->moveto_next_part()) {
      $self->screen->index($self->{navigate}->{node_index});
      $self->speechs->index($self->{navigate}->{node_index});
#      $self->speech($self->screen->lines->[$self->screen->index]);
    } else {
      $self->speech(gettext('No more parts'));
    }
  }

  # left
  elsif ($key eq $c->KEYS_LEFT) {
    my $browser = $self->{browser};
    my $html = $browser->current_tab->pop_control();
    if ($html) {
      $html->screen->update(1);
      $browser->current_tab->current_control($html);
      $self->detach();
    }
  }

  # execute elements
  elsif ($key eq $c->KEYS_RIGHT || $key eq "\n") {
    my $node = $self->{navigate}->index_node;

    # link
    if ($node->tag eq 'a') {
      my $url = $node->attr('href');
      $url = decode($self->{encoding}, $url);
      if ($url =~ /^#(.*)/) {
	# ...
      } else {
	my $browser = $self->{browser};
	$url = $self->get_abs_url($url);
	$browser->current_tab->push_control($self);
	$browser->open_page($url);
      }
    }

    # frame, iframe
    elsif ($node->tag eq 'frame'
	|| $node->tag eq 'iframe') {
      my $url = $node->attr('src');
      my $browser = $self->{browser};
      $self->detach();
      $browser->open_page($url);
    }

    # img longdesc
    elsif ($node->tag eq 'img') {
    }

    # object data='*.html'
    elsif ($node->tag eq 'object') {
    }

    # img usemap
    elsif ($node->tag eq 'img'
	   && $node->attr('usemap')) {
}

    # select
    elsif ($node->tag eq 'select') {
    }

    # data table
    elsif ($node->tag eq 'div'
	   && $node->attr('class') eq 'e-guidedog_table') {
    }

    # input text
    elsif ($node->tag eq 'input'
	   && ((not defined $node->attr('type')) || lc($node->attr('type')) eq 'text')) {
      my $callback = sub {
	# $self eGuideDog::Browser::HTML
	my $text = shift;
	$node->attr('value', $text);
	$self->screen->current_line($self->screen->index . ": <input type=text> $text");
	$self->speechs->current_line("Text input: $text");
	$self->{browser}->current_tab->current_control($self);
	$self->screen->update(1);
      };
      my $input = eGuideDog::Browser::InputText->new($callback);
      $self->{browser}->current_tab->current_control($input);
    }

    # input password
    elsif ($node->tag eq 'input'
	   && (lc($node->attr('type')) eq 'password')) {
      my $callback = sub {
	# $self eGuideDog::Browser::HTML
	my $text = shift;
	$node->attr('value', $text);
	$self->screen->current_line($self->screen->index . ": <input type=password> " . ('*' x length($text)));
	$self->speechs->current_line("Text input: " . ('*' x length($text)));
	$self->{browser}->current_tab->current_control($self);
	$self->screen->update(1);
      };
      my $input = eGuideDog::Browser::InputPassword->new($callback);
      $self->{browser}->current_tab->current_control($input);
    }

    # input textarea
    # input image

    # input submit
    elsif (($node->tag eq 'input'
	    && (lc($node->attr('type')) eq 'submit'))
	   || ($node->tag eq 'button'
	      && ($node->attr('type')
		  && (lc($node->attr('type')) eq 'submit')))) {
      # find form node
      my @form_nodes = $node->look_up('_tag', 'form');
      if (!$form_nodes[0]) {
	warn('submit button without form');
	return;
      }
      my $method = lc($form_nodes[0]->attr('method'));
      $method = 'get' unless $method; # default method is get
      my $action = $form_nodes[0]->attr('action');
      $action = $self->get_abs_url($action);
      my $enctype = $form_nodes[0]->attr('enctype');
      $enctype = 'application/x-www-form-urlencoded' unless $enctype;

      # generate content
      my @contents;
      my $content;
      $self->gen_submit_content($form_nodes[0], \@contents);
      my $name = shift(@contents);
      my $value = shift(@contents);
      if ($name) {
	$content = $name . '=' . $value;
	while ($name = shift(@contents)) {
	  $content .= '&' . $name . '=' . shift(@contents);
	}
      }

      # create request
      if ($method eq 'post') {
	my $browser = $self->{browser};
	$browser->current_tab->push_control($self);
	$browser->open_page($action, $content);	
      } else { # $method eq 'get'
	my $browser = $self->{browser};
	$browser->current_tab->push_control($self);
	$browser->open_page($action . '?' . $content);	
      }
    }

    # checkbox
    # radio
    # reset
    # file
    # button
  }

  # tab
  elsif ($key eq "\t") {
    my $found = 0;
    my $v = $self->{navigate};
    if ($v->{tabindex}) {
      my $len = @{$v->{tabindex}};
      my $i;
      for ($i = 0; $i < $len; $i++) {
	last if ($v->{tabindex}->[$i] == $v->{current_node});
      }
      $i++;
      $i = 0 if ($i >= $len);
      $v->restart();
      while ($v->moveto_next_valid_node()) {
	$v->{node_index}++;
	last if (ref $v->{current_node} && $v->{current_node} == $v->{tabindex}->[$i]);
      }
    } else { # no tabindex, find naturally next
      while ($v->moveto_next_valid_node()) {
	$v->{node_index}++;
	if (ref $v->{current_node}
	    && ($v->{current_node}->tag eq 'a'
		|| $v->{current_node}->tag eq 'area'
		|| $v->{current_node}->tag eq 'button'
		|| $v->{current_node}->tag eq 'input'
		|| $v->{current_node}->tag eq 'label'
		|| $v->{current_node}->tag eq 'legend'
		|| $v->{current_node}->tag eq 'textarea')) {
	  $found = 1;
	  last;
	}
      }
      if (!$found) {
	$v->restart();
	while ($v->moveto_next_valid_node()) {
	  $v->{node_index}++;
	  if (ref $v->{current_node}
	      && ($v->{current_node}->tag eq 'a'
		  || $v->{current_node}->tag eq 'area'
		  || $v->{current_node}->tag eq 'button'
		  || $v->{current_node}->tag eq 'input'
		  || $v->{current_node}->tag eq 'label'
		  || $v->{current_node}->tag eq 'legend'
		  || $v->{current_node}->tag eq 'textarea')) {
	    $found = 1;
	    last;
	  }
	}
      }

      if (!$found) {
	$v->restart();
	$self->screen->index(0);
	$self->speechs->index(0);
      }
    }
  }

  # unknown function key
  else {
    return eGuideDog::Browser::Const::THROW_KEY;
  }

  return eGuideDog::Browser::Const::CATCH_KEY;
}

1;
